﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Interbotix_Robot_Turret_Demo
{
    public partial class RoboTurretDemo : Form
    {
        System.IO.Ports.SerialPort port = new System.IO.Ports.SerialPort();
        Dictionary<string, byte> bytepacketsend = new Dictionary<string, byte>();
        Dictionary<string, int> pantilt1023 = new Dictionary<string, int>();
        RR_COM_APILib.API_Wrapper RR = new RR_COM_APILib.API_Wrapper();

        int intCameraPixelWidth = 0;
        int intCameraPixelHeight = 0;
        int intCenterPan = 0;
        int intCenterTilt = 0;
        int intOffsetPan = 0;
        int intOffsetTilt = 0;
        int intTempPan = 0;
        int intTempTilt = 0;
        int cogX = 0;
        int cogY = 0;
        int intBlobSizeMinimum = 0;
        int intFactor = 0;

        public RoboTurretDemo()
        {
            InitializeComponent();
        }

#region "Form Open & Close"

        private void FormDemo_Load(object sender, EventArgs e)
        {
            //Populate the com port drop down
            foreach (string port in System.IO.Ports.SerialPort.GetPortNames())
            {
                cmbPorts.Items.Add(port);
                cmbPorts.Text = port;
            }

            //populate our listbox
            lstColor.Items.Add("RED");
            lstColor.SelectedIndex = 0;
            lstColor.Items.Add("YELLOW");
            lstColor.Items.Add("BLUE");
            lstColor.Items.Add("GREEN");

            //sending byte packet
            bytepacketsend.Add("header_byte", Convert.ToByte(255));
            // always begin packet with 255 header byte - static, leave alone
            bytepacketsend.Add("panH", Convert.ToByte(1));
            bytepacketsend.Add("panL", Convert.ToByte(254));
            bytepacketsend.Add("tiltH", Convert.ToByte(1));
            bytepacketsend.Add("tiltL", Convert.ToByte(255));
            bytepacketsend.Add("buttons", Convert.ToByte(0));
            bytepacketsend.Add("ext_instructions", Convert.ToByte(8));
            bytepacketsend.Add("checksum", Convert.ToByte((255 - (Convert.ToInt32(bytepacketsend["panL"]) + Convert.ToInt32(bytepacketsend["panH"]) + Convert.ToInt32(bytepacketsend["tiltL"]) + Convert.ToInt32(bytepacketsend["tiltH"]) + Convert.ToInt32(bytepacketsend["buttons"]) + Convert.ToInt32(bytepacketsend["ext_instructions"])) % 256)));

            //we use these to store a 0 to 1023 value from the scrollbars to make life easy when drawing the crosshairs in pbAim_Paint
            pantilt1023.Add("pan", 512);
            pantilt1023.Add("tilt", 512);

            intBlobSizeMinimum = 20;
            tbBlobSizeMin.Value = intBlobSizeMinimum;
            txtBlobSizeMin.Text = intBlobSizeMinimum.ToString();

            //intFactor is the divider we use to translate pixel movement into servo movement.
            //Good settings for the different dynamixels are as follows.
            //AX-12 = 10
            //AX-18 = 7
            intFactor = 10;
            tbFactor.Value = intFactor;
            txtFactor.Text = intFactor.ToString();
        }

        private void FormDemo_FormClosing(object sender, FormClosingEventArgs e)
        {
            // if the port is still open when the form is closing, be sure to default all values and close the port.
            if (port.IsOpen)
            {
                try
                {
                    ResetDefaults();
                    port.Close();
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error updating values: " + ex.Message);
                }
            }
        }

        private void ResetDefaults()
        {
            bytepacketsend["header_byte"] = Convert.ToByte(255);
            bytepacketsend["panH"] = Convert.ToByte(1);
            bytepacketsend["panL"] = Convert.ToByte(255);
            bytepacketsend["tiltH"] = Convert.ToByte(1);
            bytepacketsend["tiltL"] = Convert.ToByte(255);
            bytepacketsend["buttons"] = Convert.ToByte(0);
            bytepacketsend["ext_instructions"] = Convert.ToByte(8);
            bytepacketsend["checksum"] = Convert.ToByte((255 - (Convert.ToInt32(bytepacketsend["panL"]) + Convert.ToInt32(bytepacketsend["panH"]) + Convert.ToInt32(bytepacketsend["tiltL"]) + Convert.ToInt32(bytepacketsend["tiltH"]) + Convert.ToInt32(bytepacketsend["buttons"]) + Convert.ToInt32(bytepacketsend["ext_instructions"])) % 256));

            pantilt1023["pan"] = 512;
            pantilt1023["tilt"] = 512;
            hsbPan.Value = 512;
            vsbTilt.Value = 512;
            pbAim.Refresh();
            UpdateValues();

        }


#endregion
        
#region "Serial Communication

        private void UpdateValues()
        {
            txtServoPan.Text = pantilt1023["pan"].ToString();
            txtServoTilt.Text = pantilt1023["tilt"].ToString();

            //Do our checksum at the last moment to make sure it is up to date
            bytepacketsend["checksum"] = Convert.ToByte((255 - (Convert.ToInt32(bytepacketsend["panH"]) + Convert.ToInt32(bytepacketsend["panL"]) + Convert.ToInt32(bytepacketsend["tiltH"]) + Convert.ToInt32(bytepacketsend["tiltL"]) + Convert.ToInt32(bytepacketsend["buttons"]) + Convert.ToInt32(bytepacketsend["ext_instructions"])) % 256));

            if (port.IsOpen)
            {
                try
                {
                    port.Write(bytepacketsend.Values.ToArray(), 0, bytepacketsend.Count());
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error updating values: " + ex.Message);
                }
            }

        }

        private void btnSetPort_Click(object sender, EventArgs e)
        {
            try
            {
                if (!port.IsOpen)
                {
                    // Setup SerialPort
                    port.PortName = cmbPorts.Text;
                    port.BaudRate = 38400;
                    port.DataBits = 8;
                    port.StopBits = System.IO.Ports.StopBits.One;
                    port.Parity = System.IO.Ports.Parity.None;
                    port.Open();
                    hsbPan.Enabled = true;
                    vsbTilt.Enabled = true;
                    btnSetPort.Text = "Close Port";
                    UpdateValues();
                }
                else
                {
                    ResetDefaults();
                    port.Close();
                    hsbPan.Enabled = false;
                    vsbTilt.Enabled = false;
                    btnSetPort.Text = "Open Port";
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show("Communications Error: " + ex.Message);
            }
        }

#endregion

#region "Pan & Tilt Controls"

        private void Scrollbars(object sender, ScrollEventArgs e)
        {
            fncCalculateBytes(1023 - hsbPan.Value, 1023 - vsbTilt.Value);        
            pbAim.Refresh();
            UpdateValues();
        }

        private void pbAim_MouseMove(object sender, MouseEventArgs e)
        {
            // only move the servos when the left mouse button is clicked,
            // the port is open, and the location of the mouse is within the correct intBuffer.
            // the box is 512 x 512 since 1024 x 1024 would be HUGE on the form.
            if (e.Button == MouseButtons.Left && port.IsOpen)
            {
                if (e.Y > 0 && e.Y <= 512 && e.X > 0 && e.X <= 512)
                {
                    //we do "(e.X * 2) - 1" because the picturebox is half size at 512x512. We need to convert that over to 0 to 1023.
                    //we actually are only getting 1 to 1023, but if we can live with that so can you.
                    fncCalculateBytes(1023 - ((e.X * 2) - 1), 1023 - ((e.Y * 2) - 1));
                    hsbPan.Value = (e.X * 2) - 1;
                    vsbTilt.Value = (e.Y * 2) - 1;
                    pbAim.Refresh();
                    UpdateValues();
                }
            }
        }

        private void fncCalculateBytes(int intPan, int intTilt)
        {
            //for the servo horn 1023 is fully rotated left / 0 is fully rotated right
            //thus tilt going up is the # growing toward 1023
            //and panning left (CCW) is the # growing toward 1023

            double r = 0;
            int i = 0;

            //store the 0 - 1023 values
            pantilt1023["pan"] = Convert.ToInt32(intPan);
            pantilt1023["tilt"] = Convert.ToInt32(intTilt);

            //AX & RX Dynamixels have a resolution of 1024 steps (0 to 1023) over 300 degrees
            //But a byte can only store a max # of 255 so we have to do some math magic to split the information into 2 bytes


            // i = Math.Truncate(intPan / 256)); Cannot use this for some reason???

            i = (int)(intPan / 256);
            r = intPan - (i * 256);
            bytepacketsend["panL"] = Convert.ToByte(r);
            bytepacketsend["panH"] = Convert.ToByte(i);

            // i = Math.Truncate(intTilt / 256); Cannot use this for some reason???
            
            i = (int)(intTilt / 256);
            r = intTilt - (i * 256);
            bytepacketsend["tiltL"] = Convert.ToByte(r);
            bytepacketsend["tiltH"] = Convert.ToByte(i);

        }

        private void pbAim_MouseUp(object sender, MouseEventArgs e)
        {
            pbAim.Refresh();
        }

        private void pbAim_Paint(object sender, PaintEventArgs e)
        {
            // create a graphics object on the PictureBox and draw two crosshair lines with it.
            Graphics gfx = e.Graphics;
            Pen myPen = new Pen(Color.Black);

            gfx.DrawLine(myPen, new Point((1023 - pantilt1023["pan"]) / 2, 0), new Point((1023 - pantilt1023["pan"]) / 2, pbAim.Height));
            //Draw verticle line
            gfx.DrawLine(myPen, new Point(0, (1023 - pantilt1023["tilt"]) / 2), new Point(pbAim.Width, (1023 - pantilt1023["tilt"]) / 2));
            //Draw horizontal line
        }

#endregion

#region "RoboRealm Code"

        private void btnRR_Click(object sender, EventArgs e)
        {
            if (btnRR.Text == "Initialize RoboRealm")
            {
                object test = 6060; 

                RR.Open("C:\\Program Files\\RoboRealm\\RoboRealm.exe", ref test);
                RR.SetCamera("on");

                if (!Convert.ToBoolean(RR.Connect("localhost")))
                {
                    //Make sure you have turned on the RoboRealm API!!!!!!! OPTIONS > API SERVER TAB > ACTIVATE CHECKBOX
                    MessageBox.Show("Could not connect to RoboRealm", "Connection Error", MessageBoxButtons.OK);
                    return;
                }
                else
                {
                    //Make sure you have preset your camera resolution in RoboRealm 
                    //(OPTIONS > VIDEO FORMAT > OUTPUT SIZE > 320 x 240)
                    //the demo code is written to work with 320x240 (less pixels = faster response)
                    intCameraPixelWidth = Convert.ToInt32(RR.GetVariable("IMAGE_WIDTH"));
                    intCameraPixelHeight = Convert.ToInt32(RR.GetVariable("IMAGE_HEIGHT"));
                    intCenterPan = intCameraPixelWidth / 2;
                    intCenterTilt = intCameraPixelHeight / 2;
                    Set_Filters();
                    tmrTracking.Enabled = true;
                    btnRR.Text = "Close RoboRealm";
                    btnRR.ForeColor = Color.Red;
                }
            }
            else
            {
                tmrTracking.Enabled = false;
                RR.Close();
                chkTracking.Checked = false;
                txtBlobSizeMin.BackColor = Color.AliceBlue;
                btnRR.Text = "Initialize RoboRealm";
                btnRR.ForeColor = Color.Black;
            }
        }

        private void tmrTracking_Tick(object sender, EventArgs e)
        {
            txtBlobSize.Text = RR.GetVariable("COG_BOX_SIZE");


            if (Convert.ToInt32(RR.GetVariable("COG_BOX_SIZE")) > intBlobSizeMinimum)
            {
                txtBlobSize.BackColor = Color.LightGreen;
                intTempPan = pantilt1023["pan"];
                intTempTilt = pantilt1023["tilt"];

                //grab the center of gravity of the tracked object
                cogX = Convert.ToInt32(RR.GetVariable("COG_X"));
                cogY = Convert.ToInt32(RR.GetVariable("COG_Y"));
                txtCOGPan.Text = cogX.ToString();
                txtCOGTilt.Text = cogY.ToString();

                //Calculate the offset from center of camera vision
                intOffsetPan = cogX - intCenterPan;
                intOffsetTilt = cogY - intCenterTilt;
                txtOffsetPan.Text = intOffsetPan.ToString();
                txtOffsetTilt.Text = intOffsetTilt.ToString();

                //apply factor to offset and add to current position
                intTempPan = intTempPan - (intOffsetPan / intFactor);
                //7 is a factor for reducing movement to smooth tracking
                intTempTilt = intTempTilt + (intOffsetTilt / intFactor);
                //7 is a factor for reducing movement to smooth tracking

                //Stay inside 1 to 1023
                if (intTempPan > 1023)
                    intTempPan = 1023;
                if (intTempPan < 1)
                    intTempPan = 1;
                if (intTempTilt > 1023)
                    intTempTilt = 1023;
                if (intTempTilt < 1)
                    intTempTilt = 1;

                if (chkTracking.Checked)
                {
                    //write new position to Pan & Tilt
                    fncCalculateBytes(intTempPan, intTempTilt);
                    //Update form
                    pbAim.Refresh();
                    hsbPan.Value = 1023 - pantilt1023["pan"];
                    vsbTilt.Value = 1023 - pantilt1023["tilt"];
                    //send to microcontroller
                    UpdateValues();
                }
            }
            else
            {
                txtBlobSize.BackColor = Color.AliceBlue;
            }
        }

        private void lstColor_SelectedIndexChanged(object sender, EventArgs e)
        {
            Set_Filters();
        }

        private void Set_Filters()
        {
            //The predefined filters will most likely need to be adjusted for your specific envirionment.
            //The settings need to be adjusted for specific cameras, lighting, and object color tones.
            //Simply double click the .robo file to open it in RoboRealm then make adjustments using the 
            //interface and re-save the files.
            switch (lstColor.SelectedItem.ToString())
            {
                case "RED":
                    RR.LoadProgram(Application.StartupPath + "\\robofiles\\MS_lifecam_Cinema_track_Red.robo");
                    break;
                case "YELLOW":
                    RR.LoadProgram(Application.StartupPath + "\\robofiles\\MS_lifecam_Cinema_track_Yellow.robo");
                    break;
                case "BLUE":
                    RR.LoadProgram(Application.StartupPath + "\\robofiles\\MS_lifecam_Cinema_track_Blue.robo");
                    break;
                case "GREEN":
                    RR.LoadProgram(Application.StartupPath + "\\robofiles\\MS_lifecam_Cinema_track_Green.robo");
                    break;
            }
        }

        private void tbBlobSizeMin_Scroll(object sender, EventArgs e)
        {                    
            intBlobSizeMinimum = tbBlobSizeMin.Value;
            txtBlobSizeMin.Text = tbBlobSizeMin.Value.ToString();
        }

        private void tbFactor_Scroll(object sender, EventArgs e)
        {                   
            intFactor = tbFactor.Value;
            txtFactor.Text = intFactor.ToString();
        }

#endregion

    }
}
